import { TileID, polygon2buffer, BaseLayer, LayerSourceType, RenderType } from 'wind-gl-core';
export { DecodeType, ImageSource, LayerSourceType, MaskType, RenderFrom, RenderType, TileID, TileSource, TimelineSource, configDeps } from 'wind-gl-core';
import maplibregl from 'maplibre-gl';
import rewind from '@mapbox/geojson-rewind';
import { highPrecision, Matrix4, OrthographicCamera, PerspectiveCamera, Vector3, ProjectionMatrix, utils, Renderer, Scene } from '@sakitam-gis/vis-engine';
import RBush from '@sakitam-gis/rbush';

const earthRadius = 63710088e-1;
const earthCircumference = 2 * Math.PI * earthRadius;
function circumferenceAtLatitude(latitude) {
  return earthCircumference * Math.cos(latitude * Math.PI / 180);
}
function mercatorZfromAltitude(altitude, lat) {
  return altitude / circumferenceAtLatitude(lat);
}
function lngFromMercatorX(x, wrap = 0) {
  return x * 360 - 180 + wrap * 360;
}
function latFromMercatorY(y) {
  const y2 = 180 - y * 360;
  return 360 / Math.PI * Math.atan(Math.exp(y2 * Math.PI / 180)) - 90;
}
function getCoordinatesCenterTileID(coords) {
  let minX = Infinity;
  let minY = Infinity;
  let maxX = -Infinity;
  let maxY = -Infinity;
  for (const coord of coords) {
    minX = Math.min(minX, coord.x);
    minY = Math.min(minY, coord.y);
    maxX = Math.max(maxX, coord.x);
    maxY = Math.max(maxY, coord.y);
  }
  const dx = maxX - minX;
  const dy = maxY - minY;
  const dMax = Math.max(dx, dy);
  const zoom = Math.max(0, Math.floor(-Math.log(dMax) / Math.LN2));
  const tilesAtZoom = Math.pow(2, zoom);
  return {
    z: zoom,
    x: Math.floor((minX + maxX) / 2 * tilesAtZoom),
    y: Math.floor((minY + maxY) / 2 * tilesAtZoom),
    extent: [minX, minY, maxX, maxY]
  };
}

const { degToRad, radToDeg, clamp } = utils;
highPrecision(true);
class CameraSync {
  constructor(map, cameraType, scene) {
    this.worldMatrix = new Matrix4();
    this.mercatorMatrix = new Matrix4();
    this.labelPlaneMatrix = new Matrix4();
    this.glCoordMatrix = new Matrix4();
    const { width, height } = map.transform;
    const fov = radToDeg(Math.atan(3 / 4));
    const nearZ = 0.1;
    const farZ = 1e21;
    this.map = map;
    this.scene = scene;
    this.scene.matrixAutoUpdate = false;
    this.scene.worldMatrixNeedsUpdate = true;
    this.camera = cameraType === "orthographic" ? new OrthographicCamera(-width / 2, width / 2, height / 2, -height / 2, nearZ, farZ) : new PerspectiveCamera(fov, width / height, nearZ, farZ);
    this.camera.matrixAutoUpdate = false;
    this.camera.position.z = 600;
    this.setup();
  }
  setup() {
    const { width, height } = this.map.transform;
    const maxPitch = degToRad(this.map.transform.maxPitch);
    this.camera.aspect = width / height;
    this.halfFov = this.map.transform.fov / 2;
    this.cameraToCenterDistance = 0.5 / Math.tan(this.halfFov) * height;
    this.acuteAngle = Math.PI / 2 - maxPitch;
    this.update();
  }
  update() {
    const { width, height, worldSize, elevation, _camera } = this.map.transform;
    const center = this.map.getCenter();
    const pitch = this.map.getPitch();
    const pitchRad = degToRad(pitch);
    const bearing = this.map.getBearing();
    const fovRad = this.map.transform._fov;
    const halfFov = fovRad / 2;
    const pitchAngle = Math.cos(Math.PI / 2 - pitchRad);
    const groundAngle = Math.PI / 2 + pitchRad;
    this.cameraToCenterDistance = 0.5 / Math.tan(halfFov) * height;
    const point = this.map.transform.project(center);
    const rotateMap = new Matrix4().fromRotationZ(Math.PI);
    const scale = new Matrix4().fromScale(new Vector3(-worldSize, worldSize, worldSize));
    const translateMap = new Matrix4().fromTranslation(new Vector3(-point.x, point.y, 0));
    const nz = height / 50;
    const nearZ = Math.max(nz * pitchAngle, nz);
    const fovAboveCenter = fovRad * (0.5 + this.map.transform.centerOffset.y / height);
    const topHalfSurfaceDistance = Math.sin(fovAboveCenter) * this.cameraToCenterDistance / Math.sin(clamp(Math.PI - groundAngle - fovAboveCenter, 0.01, Math.PI - 0.01));
    const pixelsPerMeter = mercatorZfromAltitude(1, center.lat) * worldSize || 1;
    const furthestDistance = pitchAngle * topHalfSurfaceDistance + this.cameraToCenterDistance;
    const farZ = furthestDistance * 1.01;
    this.mercatorMatrix = new Matrix4().scale(new Vector3(worldSize, worldSize, worldSize / pixelsPerMeter));
    const may = new Matrix4().fromTranslation(new Vector3(0, 0, this.cameraToCenterDistance));
    this.labelPlaneMatrix = new Matrix4();
    const m = new Matrix4();
    m.scale(new Vector3(1, -1, 1));
    m.translate(new Vector3(-1, -1, 0));
    m.scale(new Vector3(2 / width, 2 / height, 1));
    this.glCoordMatrix = m;
    this.camera.aspect = width / height;
    this.cameraTranslateZ = this.cameraToCenterDistance;
    if (this.camera instanceof OrthographicCamera) {
      this.camera.projectionMatrix.orthographic(-width / 2, width / 2, height / 2, -height / 2, nearZ, farZ);
    } else {
      this.camera.projectionMatrix.perspective(fovRad, width / height, nearZ, farZ);
    }
    const cameraWorldMatrix = new Matrix4().premultiply(may).premultiply(new Matrix4().fromRotationX(pitchRad)).premultiply(new Matrix4().fromRotationZ(-degToRad(bearing)));
    if (elevation)
      cameraWorldMatrix.elements[14] = _camera.position[2] * worldSize;
    this.camera.worldMatrix.copy(cameraWorldMatrix);
    this.camera.updateMatrixWorld();
    if (this.scene) {
      this.scene.localMatrix = new ProjectionMatrix().premultiply(rotateMap).premultiply(scale).premultiply(translateMap);
    }
  }
}

function getTileProjBounds(tileID) {
  const numTiles = 1 << tileID.z;
  return {
    left: tileID.wrapedX / numTiles,
    top: tileID.wrapedY / numTiles,
    right: (tileID.wrapedX + 1) / numTiles,
    bottom: (tileID.wrapedY + 1) / numTiles
  };
}
function getTileBounds(tileID) {
  const { z, x, y } = tileID;
  const wrap = tileID.wrap;
  const numTiles = 1 << z;
  const leftLng = lngFromMercatorX(x / numTiles, wrap);
  const rightLng = lngFromMercatorX((x + 1) / numTiles, wrap);
  const topLat = latFromMercatorY(y / numTiles);
  const bottomLat = latFromMercatorY((y + 1) / numTiles);
  return [leftLng, bottomLat, rightLng, topLat];
}
function intersects(a, b) {
  return b.minX < a.maxX && b.minY < a.maxY && b.maxX > a.minX && b.maxY > a.minY;
}
function expandTiles(tiles) {
  if (!tiles || tiles.length === 0)
    return [];
  const level = /* @__PURE__ */ new Map();
  const tree = new RBush(9);
  for (let i = 0; i < tiles.length; i++) {
    const tileID = tiles[i];
    const { z, wrapedX, wrapedY } = tileID;
    const cache = level.get(z);
    if (!cache) {
      level.set(z, [tileID]);
    } else {
      level.set(z, [...cache, tileID]);
    }
  }
  const object = {};
  for (const [lk, lv] of level) {
    let xmin2 = Infinity;
    let ymin2 = Infinity;
    let xmax2 = -Infinity;
    let ymax2 = -Infinity;
    let wrapMin = Infinity;
    for (let j = 0; j < lv.length; j++) {
      const litem = lv[j];
      const item = {
        minX: litem.wrapedX,
        minY: litem.wrapedY,
        maxX: litem.wrapedX + 1,
        maxY: litem.wrapedY + 1,
        tileKey: litem.tileKey
      };
      tree.insert(item);
      xmin2 = Math.min(item.minX, xmin2);
      xmax2 = Math.max(item.maxX, xmax2);
      ymin2 = Math.min(item.minY, ymin2);
      ymax2 = Math.max(item.maxY, ymax2);
      wrapMin = Math.min(litem.wrap, wrapMin);
    }
    let baseTileID = lv.find((t) => t.wrapedX === xmin2 && t.wrapedY === ymin2 && t.z === lk);
    if (!baseTileID) {
      const max = Math.pow(2, lk);
      baseTileID = new TileID(lk, wrapMin, lk, xmin2 - max * wrapMin, ymin2, lv[0].options);
    }
    object[lk] = {
      baseTileID,
      config: [xmin2, ymin2, xmax2, ymax2]
    };
  }
  const keys = Object.keys(object).sort().reverse().map(Number);
  const addTiles = [];
  for (let k = 0; k < keys.length; k++) {
    const { config, baseTileID } = object[keys[k]];
    const xd = config[2] - config[0];
    const yd = config[3] - config[1];
    for (let x = 0; x < xd; x++) {
      for (let y = 0; y < yd; y++) {
        const tile = baseTileID.neighbor(x, y);
        if (!tiles.find((t) => t.tileKey === tile.tileKey)) {
          const result = tree.collides(
            {
              minX: tile.wrapedX,
              minY: tile.wrapedY,
              maxX: tile.wrapedX + 1,
              maxY: tile.wrapedY + 1
            },
            { intersects }
          );
          if (!result) {
            addTiles.push(tile);
          }
        }
      }
    }
  }
  return addTiles;
}

class Layer {
  constructor(id, source, options) {
    this.id = id;
    this.type = "custom";
    this.renderingMode = options?.renderingMode || "2d";
    this.options = {
      ...options || {}
    };
    this.source = source;
    this.update = this.update.bind(this);
    this.moveStart = this.moveStart.bind(this);
    this.moveEnd = this.moveEnd.bind(this);
    this.handleZoom = this.handleZoom.bind(this);
    this.handleResize = this.handleResize.bind(this);
  }
  get camera() {
    return this.sync.camera;
  }
  update() {
    this.sync.update();
    if (this.layer) {
      this.layer.update();
    }
  }
  moveStart() {
    if (this.layer) {
      this.layer.moveStart();
    }
  }
  moveEnd() {
    if (this.layer) {
      this.layer.moveEnd();
    }
  }
  handleResize() {
    if (this.renderer && this.gl) {
      const { width, height } = this.map.painter;
      this.renderer.setSize(width, height);
      if (this.layer) {
        this.layer.resize(width, height);
      }
      this.update();
    }
  }
  handleZoom() {
    if (this.layer) {
      this.layer.handleZoom();
    }
  }
  updateOptions(options) {
    this.options = {
      ...this.options,
      ...options || {}
    };
    if (this.layer) {
      this.layer.updateOptions(options);
    }
  }
  getMask() {
    return this.options.mask;
  }
  processMask() {
    if (this.options.mask) {
      const mask = this.options.mask;
      const data = mask.data;
      rewind(data, true);
      const tr = (coords) => {
        const mercatorCoordinates = [];
        for (let i2 = 0; i2 < coords.length; i2++) {
          const coord = coords[i2];
          const p = maplibregl.MercatorCoordinate.fromLngLat(coord);
          mercatorCoordinates.push([p.x, p.y]);
        }
        return mercatorCoordinates;
      };
      const features = data.features;
      const len = features.length;
      let i = 0;
      const fs = [];
      for (; i < len; i++) {
        const feature = features[i];
        const coordinates = feature.geometry.coordinates;
        const type = feature.geometry.type;
        if (type === "Polygon") {
          fs.push({
            type: "Feature",
            properties: {},
            geometry: {
              type: "Polygon",
              coordinates: feature.geometry.coordinates.map((c) => tr(c))
            }
          });
        } else if (type === "MultiPolygon") {
          const css = [];
          for (let k = 0; k < coordinates.length; k++) {
            const coordinate = coordinates[k];
            const cs = [];
            for (let n = 0; n < coordinate.length; n++) {
              cs.push(tr(coordinates[k][n]));
            }
            css.push(cs);
          }
          fs.push({
            type: "Feature",
            properties: {},
            geometry: {
              type: "MultiPolygon",
              coordinates: css
            }
          });
        }
      }
      return {
        data: polygon2buffer(fs),
        type: mask.type
      };
    }
  }
  setMask(mask) {
    this.options.mask = Object.assign({}, this.options.mask, mask);
    if (this.layer) {
      this.layer.setMask(this.processMask());
    }
  }
  onAdd(m, gl) {
    this.gl = gl;
    this.map = m;
    const canvas = m.getCanvas();
    this.renderer = new Renderer(gl, {
      autoClear: false,
      extensions: [
        "OES_texture_float",
        "OES_texture_float_linear",
        "WEBGL_color_buffer_float",
        "EXT_color_buffer_float"
      ]
    });
    this.scene = new Scene();
    this.sync = new CameraSync(this.map, "perspective", this.scene);
    this.planeCamera = new OrthographicCamera(0, 1, 1, 0, 0, 1);
    this.layer = new BaseLayer(
      this.source,
      {
        renderer: this.renderer,
        scene: this.scene
      },
      {
        renderType: this.options.renderType,
        renderFrom: this.options.renderFrom,
        styleSpec: this.options.styleSpec,
        displayRange: this.options.displayRange,
        widthSegments: this.options.widthSegments,
        heightSegments: this.options.heightSegments,
        wireframe: this.options.wireframe,
        picking: this.options.picking,
        mask: this.processMask(),
        getZoom: () => this.map?.getZoom(),
        triggerRepaint: () => {
          this.map?.triggerRepaint();
        },
        getTileProjSize: (z) => {
          const w = 1 / Math.pow(2, z);
          return [w, w];
        },
        getPixelsToUnits: () => {
          const pixel = 1;
          const y = canvas.clientHeight / 2 - pixel / 2;
          const x = canvas.clientWidth / 2 - pixel / 2;
          const left = maplibregl.MercatorCoordinate.fromLngLat(m.unproject([x, y]));
          const right = maplibregl.MercatorCoordinate.fromLngLat(m.unproject([x + pixel, y + pixel]));
          return [Math.abs(right.x - left.x), Math.abs(left.y - right.y)];
        },
        getPixelsToProjUnit: () => [
          this.map?.transform.pixelsPerMeter,
          this.map?.transform.pixelsPerMeter
        ],
        getViewTiles: (source, renderType) => {
          let { type } = source;
          type = type !== LayerSourceType.timeline ? type : source.privateType;
          const map = this.map;
          if (!map)
            return [];
          const { transform } = map;
          const wrapTiles = [];
          if (type === LayerSourceType.image) {
            const cornerCoords = source.coordinates.map((c) => maplibregl.MercatorCoordinate.fromLngLat(c));
            const tileID = getCoordinatesCenterTileID(cornerCoords);
            if (source.wrapX) {
              transform.getVisibleUnwrappedCoordinates(tileID).forEach((unwrapped) => {
                const { canonical, wrap } = unwrapped;
                const { x, y, z } = canonical;
                wrapTiles.push(
                  new TileID(z, wrap, z, x, y, {
                    getTileBounds: () => [
                      source.coordinates[0][0],
                      source.coordinates[2][1],
                      source.coordinates[1][0],
                      source.coordinates[0][1]
                    ],
                    getTileProjBounds: () => ({
                      left: tileID.extent[0] + wrap,
                      top: tileID.extent[1],
                      right: tileID.extent[2] + wrap,
                      bottom: tileID.extent[3]
                    })
                  })
                );
              });
            } else {
              const x = tileID.x;
              const y = tileID.y;
              const z = tileID.z;
              const wrap = 0;
              wrapTiles.push(
                new TileID(z, wrap, z, x, y, {
                  getTileBounds: () => [
                    source.coordinates[0][0],
                    source.coordinates[2][1],
                    source.coordinates[1][0],
                    source.coordinates[0][1]
                  ],
                  getTileProjBounds: () => ({
                    left: tileID.extent[0] + wrap,
                    top: tileID.extent[1],
                    right: tileID.extent[2] + wrap,
                    bottom: tileID.extent[3]
                  })
                })
              );
            }
          } else if (type === LayerSourceType.tile) {
            const opts = {
              tileSize: utils.isNumber(this.source.tileSize) ? source.tileSize : source.tileSize?.[0] || 512,
              // for mapbox
              minzoom: source.minZoom,
              maxzoom: source.maxZoom,
              roundZoom: source.roundZoom
            };
            const tiles = transform.coveringTiles(opts);
            for (let i = 0; i < tiles.length; i++) {
              const tile = tiles[i];
              const { canonical, wrap } = tile;
              const { x, y, z } = canonical;
              if (source.wrapX) {
                wrapTiles.push(
                  new TileID(z, wrap, z, x, y, {
                    getTileBounds,
                    getTileProjBounds
                  })
                );
              } else if (tile.wrap === 0) {
                wrapTiles.push(
                  new TileID(z, wrap, z, x, y, {
                    getTileBounds,
                    getTileProjBounds
                  })
                );
              }
            }
            if (renderType === RenderType.particles) {
              wrapTiles.push(...expandTiles(wrapTiles));
            }
          }
          return wrapTiles;
        },
        getExtent: () => {
          const map = this.map;
          const bounds = map?.getBounds().toArray();
          const xmin = bounds[0][0];
          const ymin = bounds[0][1];
          const xmax = bounds[1][0];
          const ymax = bounds[1][1];
          const minY = Math.max(ymin, map.transform.latRange ? map.transform.latRange[0] : map.transform.minLat);
          const maxY = Math.min(ymax, map.transform.latRange ? map.transform.latRange[1] : map.transform.maxLat);
          const p0 = maplibregl.MercatorCoordinate.fromLngLat(new maplibregl.LngLat(xmin, maxY));
          const p1 = maplibregl.MercatorCoordinate.fromLngLat(new maplibregl.LngLat(xmax, minY));
          return [p0.x, p0.y, p1.x, p1.y];
        },
        getGridTiles: (source) => {
          const tileSize = source.tileSize;
          const wrapX = source.wrapX;
          const map = this.map;
          if (!map)
            return [];
          const { transform } = map;
          const opts = {
            tileSize: tileSize ?? 256,
            minzoom: map.getMinZoom(),
            maxzoom: map.getMaxZoom(),
            roundZoom: false
          };
          const tiles = transform.coveringTiles(opts);
          const wrapTiles = [];
          for (let i = 0; i < tiles.length; i++) {
            const tile = tiles[i];
            const { canonical, wrap } = tile;
            const { x, y, z } = canonical;
            if (wrapX) {
              wrapTiles.push(
                new TileID(z, wrap, z, x, y, {
                  getTileBounds,
                  getTileProjBounds
                })
              );
            } else if (tile.wrap === 0) {
              wrapTiles.push(
                new TileID(z, wrap, z, x, y, {
                  getTileBounds,
                  getTileProjBounds
                })
              );
            }
          }
          return wrapTiles;
        }
      }
    );
    this.map.on("movestart", this.moveStart);
    this.map.on("move", this.update);
    this.map.on("moveend", this.moveEnd);
    this.map.on("zoom", this.handleZoom);
    this.map.on("zoomend", this.handleZoom);
    this.map.on("resize", this.handleResize);
    this.handleResize();
    this.update();
  }
  calcWrappedWorlds() {
    const result = [0];
    if (this.source?.wrapX) {
      const { width, height } = this.map.transform;
      const utl = this.map.transform.pointCoordinate(new maplibregl.Point(0, 0));
      const utr = this.map.transform.pointCoordinate(new maplibregl.Point(width, 0));
      const ubl = this.map.transform.pointCoordinate(new maplibregl.Point(width, height));
      const ubr = this.map.transform.pointCoordinate(new maplibregl.Point(0, height));
      const w0 = Math.floor(Math.min(utl.x, utr.x, ubl.x, ubr.x));
      const w1 = Math.floor(Math.max(utl.x, utr.x, ubl.x, ubr.x));
      const extraWorldCopy = 0;
      for (let w = w0 - extraWorldCopy; w <= w1 + extraWorldCopy; w++) {
        if (w === 0) {
          continue;
        }
        result.push(w);
      }
    }
    return result;
  }
  onRemove() {
    if (this.layer) {
      this.layer.destroy();
      this.layer = null;
    }
    this.map?.off("zoom", this.handleZoom);
    this.map?.off("zoomend", this.handleZoom);
    this.map?.off("movestart", this.moveStart);
    this.map?.off("move", this.update);
    this.map?.off("moveend", this.moveEnd);
    this.map?.off("resize", this.handleResize);
    this.map = null;
    this.gl = null;
  }
  prerender() {
    this.scene.worldMatrixNeedsUpdate = true;
    this.scene.updateMatrixWorld();
    this.camera.updateMatrixWorld();
    const worlds = this.calcWrappedWorlds();
    this.layer?.prerender({
      worlds,
      camera: this.camera,
      planeCamera: this.planeCamera
    });
  }
  async picker(coordinates) {
    if (!this.options.picking) {
      console.warn("[Layer]: please enable picking options!");
      return null;
    }
    if (!this.layer || !coordinates || !this.map) {
      console.warn("[Layer]: layer not initialized!");
      return null;
    }
    const point = this.map.project(coordinates);
    return this.layer.picker([point.x, point.y]);
  }
  render() {
    this.scene.worldMatrixNeedsUpdate = true;
    this.scene.updateMatrixWorld();
    this.camera.updateMatrixWorld();
    const worlds = this.calcWrappedWorlds();
    this.layer?.render({
      worlds,
      camera: this.camera,
      planeCamera: this.planeCamera
    });
  }
}

export { Layer };
//# sourceMappingURL=maplibre-wind.esm.js.map
